"""Tests the productionlist directive."""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx.addnodes import pending_xref
from sphinx.testing import restructuredtext
from sphinx.testing.util import SphinxTestApp, assert_node, etree_parse

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='productionlist')
def test_productionlist(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    warnings = app.warning.getvalue().split('\n')
    assert len(warnings) == 2
    assert warnings[-1] == ''
    assert (
        'Dup2.rst:4: WARNING: duplicate token description of Dup, other instance in Dup1'
        in warnings[0]
    )

    etree = etree_parse(app.outdir / 'index.html')
    nodes = list(etree.iter('ul'))
    assert len(nodes) >= 3

    ul = nodes[2]
    cases = []
    for li in list(ul):
        li_list = list(li)
        assert len(li_list) == 1
        p = li_list[0]
        assert p.tag == 'p'
        text = str(p.text).strip(' :')
        p_list = list(p)
        assert len(p_list) == 1
        a = p_list[0]
        assert a.tag == 'a'
        link = a.get('href')
        a_list = list(a)
        assert len(a_list) == 1
        code = a_list[0]
        assert code.tag == 'code'
        code_list = list(code)
        assert len(code_list) == 1
        span = code_list[0]
        assert span.tag == 'span'
        assert span.text is not None
        link_text = span.text.strip()
        cases.append((text, link, link_text))
    assert cases == [
        ('A', 'Bare.html#grammar-token-A', 'A'),
        ('B', 'Bare.html#grammar-token-B', 'B'),
        ('P1:A', 'P1.html#grammar-token-P1-A', 'P1:A'),
        ('P1:B', 'P1.html#grammar-token-P1-B', 'P1:B'),
        ('P2:A', 'P1.html#grammar-token-P1-A', 'P1:A'),
        ('P2:B', 'P2.html#grammar-token-P2-B', 'P2:B'),
        ('Explicit title A, plain', 'Bare.html#grammar-token-A', 'MyTitle'),
        ('Explicit title A, colon', 'Bare.html#grammar-token-A', 'My:Title'),
        ('Explicit title P1:A, plain', 'P1.html#grammar-token-P1-A', 'MyTitle'),
        ('Explicit title P1:A, colon', 'P1.html#grammar-token-P1-A', 'My:Title'),
        ('Tilde A', 'Bare.html#grammar-token-A', 'A'),
        ('Tilde P1:A', 'P1.html#grammar-token-P1-A', 'A'),
        ('Tilde explicit title P1:A', 'P1.html#grammar-token-P1-A', '~MyTitle'),
        ('Tilde, explicit title P1:A', 'P1.html#grammar-token-P1-A', 'MyTitle'),
        ('Dup', 'Dup2.html#grammar-token-Dup', 'Dup'),
        ('FirstLine', 'firstLineRule.html#grammar-token-FirstLine', 'FirstLine'),
        ('SecondLine', 'firstLineRule.html#grammar-token-SecondLine', 'SecondLine'),
    ]

    text = (app.outdir / 'LineContinuation.html').read_text(encoding='utf8')
    assert 'A</strong> ::= B C D    E F G' in text


@pytest.mark.sphinx('html', testroot='root')
def test_productionlist_xref(app: SphinxTestApp) -> None:
    text = """\
.. productionlist:: P2
   A: `:A` `A`
   B: `P1:B` `~P1:B`
"""
    doctree = restructuredtext.parse(app, text)
    refnodes = list(doctree.findall(pending_xref))
    assert_node(refnodes[0], pending_xref, reftarget='A')
    assert_node(refnodes[1], pending_xref, reftarget='P2:A')
    assert_node(refnodes[2], pending_xref, reftarget='P1:B')
    assert_node(refnodes[3], pending_xref, reftarget='P1:B')
    assert_node(refnodes[0], [pending_xref, nodes.literal, 'A'])
    assert_node(refnodes[1], [pending_xref, nodes.literal, 'A'])
    assert_node(refnodes[2], [pending_xref, nodes.literal, 'P1:B'])
    assert_node(refnodes[3], [pending_xref, nodes.literal, 'B'])


def test_productionlist_continuation_lines(
    make_app: Callable[..., SphinxTestApp], tmp_path: Path
) -> None:
    text = """
.. productionlist:: python-grammar
   assignment_stmt: (`target_list` "=")+ (`starred_expression` | `yield_expression`)
   target_list: `target` ("," `target`)* [","]
   target: `identifier`
         : | "(" [`target_list`] ")"
         : | "[" [`target_list`] "]"
         : | `attributeref`
         : | `subscription`
         : | `slicing`
         : | "*" `target`
"""
    (tmp_path / 'conf.py').touch()
    (tmp_path / 'index.rst').write_text(text, encoding='utf-8')

    app = make_app(buildername='text', srcdir=tmp_path)
    app.build(force_all=True)
    content = (app.outdir / 'index.txt').read_text(encoding='utf-8')
    expected = """\
   assignment_stmt ::= (target_list "=")+ (starred_expression | yield_expression)
   target_list     ::= target ("," target)* [","]
   target          ::= identifier
                       | "(" [target_list] ")"
                       | "[" [target_list] "]"
                       | attributeref
                       | subscription
                       | slicing
                       | "*" target
"""
    assert content == expected

    app = make_app(buildername='html', srcdir=tmp_path)
    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf-8')
    _, _, content = content.partition('<pre>')
    content, _, _ = content.partition('</pre>')
    expected = """
<strong id="grammar-token-python-grammar-assignment_stmt">assignment_stmt</strong> ::= (<a class="reference internal" href="#grammar-token-python-grammar-target_list"><code class="xref docutils literal notranslate"><span class="pre">target_list</span></code></a> &quot;=&quot;)+ (<code class="xref docutils literal notranslate"><span class="pre">starred_expression</span></code> | <code class="xref docutils literal notranslate"><span class="pre">yield_expression</span></code>)
<strong id="grammar-token-python-grammar-target_list">target_list</strong>     ::= <a class="reference internal" href="#grammar-token-python-grammar-target"><code class="xref docutils literal notranslate"><span class="pre">target</span></code></a> (&quot;,&quot; <a class="reference internal" href="#grammar-token-python-grammar-target"><code class="xref docutils literal notranslate"><span class="pre">target</span></code></a>)* [&quot;,&quot;]
<strong id="grammar-token-python-grammar-target">target</strong>          ::= <code class="xref docutils literal notranslate"><span class="pre">identifier</span></code>
                    | &quot;(&quot; [<a class="reference internal" href="#grammar-token-python-grammar-target_list"><code class="xref docutils literal notranslate"><span class="pre">target_list</span></code></a>] &quot;)&quot;
                    | &quot;[&quot; [<a class="reference internal" href="#grammar-token-python-grammar-target_list"><code class="xref docutils literal notranslate"><span class="pre">target_list</span></code></a>] &quot;]&quot;
                    | <code class="xref docutils literal notranslate"><span class="pre">attributeref</span></code>
                    | <code class="xref docutils literal notranslate"><span class="pre">subscription</span></code>
                    | <code class="xref docutils literal notranslate"><span class="pre">slicing</span></code>
                    | &quot;*&quot; <a class="reference internal" href="#grammar-token-python-grammar-target"><code class="xref docutils literal notranslate"><span class="pre">target</span></code></a>
"""
    assert content == expected



import sys
from pathlib import Path

source_dir = str(Path.cwd().resolve())
if source_dir not in sys.path:
    sys.path.insert(0, source_dir)
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.viewcode']
exclude_patterns = ['_build']


if 'test_linkcode' in tags:  # NoQA: F821 (tags is injected into conf.py)
    extensions.remove('sphinx.ext.viewcode')
    extensions.append('sphinx.ext.linkcode')

    def linkcode_resolve(domain, info):
        if domain == 'py':
            fn = info['module'].replace('.', '/')
            return 'http://foobar/source/%s.py' % fn
        elif domain == 'js':
            return 'http://foobar/js/' + info['fullname']
        elif domain in {'c', 'cpp'}:
            return 'http://foobar/%s/%s' % (domain, ''.join(info['names']))
        else:
            raise AssertionError



"""Test the Sphinx API for translator."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import pytest

from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from sphinx.testing.util import SphinxTestApp


@pytest.fixture(scope='module', autouse=True)
def _setup_module(rootdir: Path) -> Iterator[None]:
    saved_path = sys.path.copy()
    sys.path.insert(0, str(rootdir / 'test-api-set-translator'))
    yield
    sys.path[:] = saved_path


@pytest.mark.sphinx('html', testroot='root')
def test_html_translator(app: SphinxTestApp) -> None:
    # no set_translator()
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'HTML5Translator'


@pytest.mark.sphinx('html', testroot='api-set-translator')
def test_html_with_set_translator_for_html_(app: SphinxTestApp) -> None:
    # use set_translator()
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfHTMLTranslator'


@pytest.mark.sphinx('singlehtml', testroot='api-set-translator')
def test_singlehtml_set_translator_for_singlehtml(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfSingleHTMLTranslator'


@pytest.mark.sphinx('pickle', testroot='api-set-translator')
def test_pickle_set_translator_for_pickle(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfPickleTranslator'


@pytest.mark.sphinx('json', testroot='api-set-translator')
def test_json_set_translator_for_json(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfJsonTranslator'


@pytest.mark.sphinx('latex', testroot='api-set-translator')
def test_html_with_set_translator_for_latex(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfLaTeXTranslator'


@pytest.mark.sphinx('man', testroot='api-set-translator')
def test_html_with_set_translator_for_man(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfManualPageTranslator'


@pytest.mark.sphinx('texinfo', testroot='api-set-translator')
def test_html_with_set_translator_for_texinfo(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfTexinfoTranslator'


@pytest.mark.sphinx('text', testroot='api-set-translator')
def test_html_with_set_translator_for_text(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfTextTranslator'


@pytest.mark.sphinx('xml', testroot='api-set-translator')
def test_html_with_set_translator_for_xml(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfXMLTranslator'


@pytest.mark.sphinx('pseudoxml', testroot='api-set-translator')
def test_html_with_set_translator_for_pseudoxml(app: SphinxTestApp) -> None:
    translator_class = app.builder.get_translator_class()
    assert translator_class
    assert translator_class.__name__ == 'ConfPseudoXMLTranslator'



"""Test inventory util functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import sphinx.locale
from sphinx.testing.util import SphinxTestApp
from sphinx.util.inventory import InventoryFile, _InventoryItem

from tests.test_util.intersphinx_data import (
    INVENTORY_V1,
    INVENTORY_V2,
    INVENTORY_V2_AMBIGUOUS_TERMS,
    INVENTORY_V2_NO_VERSION,
)

if TYPE_CHECKING:
    from pathlib import Path


def test_read_inventory_v1() -> None:
    inv = InventoryFile.loads(INVENTORY_V1, uri='/util')
    assert inv['py:module', 'module'] == _InventoryItem(
        project_name='foo',
        project_version='1.0',
        uri='/util/foo.html#module-module',
        display_name='-',
    )
    assert inv['py:class', 'module.cls'] == _InventoryItem(
        project_name='foo',
        project_version='1.0',
        uri='/util/foo.html#module.cls',
        display_name='-',
    )


def test_read_inventory_v2() -> None:
    inv = InventoryFile.loads(INVENTORY_V2, uri='/util')

    assert len(inv.data['py:module']) == 2
    assert inv['py:module', 'module1'] == _InventoryItem(
        project_name='foo',
        project_version='2.0',
        uri='/util/foo.html#module-module1',
        display_name='Long Module desc',
    )
    assert inv['py:module', 'module2'] == _InventoryItem(
        project_name='foo',
        project_version='2.0',
        uri='/util/foo.html#module-module2',
        display_name='-',
    )
    assert inv['py:function', 'module1.func'].uri == ('/util/sub/foo.html#module1.func')
    assert inv['c:function', 'CFunc'].uri == '/util/cfunc.html#CFunc'
    assert inv['std:term', 'a term'].uri == '/util/glossary.html#term-a-term'
    assert inv['std:term', 'a term including:colon'].uri == (
        '/util/glossary.html#term-a-term-including-colon'
    )


def test_read_inventory_v2_not_having_version() -> None:
    inv = InventoryFile.loads(INVENTORY_V2_NO_VERSION, uri='/util')
    assert inv['py:module', 'module1'] == _InventoryItem(
        project_name='foo',
        project_version='',
        uri='/util/foo.html#module-module1',
        display_name='Long Module desc',
    )


@pytest.mark.sphinx('html', testroot='root')
def test_ambiguous_definition_warning(app: SphinxTestApp) -> None:
    InventoryFile.loads(INVENTORY_V2_AMBIGUOUS_TERMS, uri='/util')

    def _multiple_defs_notice_for(entity: str) -> str:
        return f'contains multiple definitions for {entity}'

    # was warning-level; reduced to info-level
    # See: https://github.com/sphinx-doc/sphinx/issues/12613
    mult_defs_a, mult_defs_b = (
        _multiple_defs_notice_for('std:term:a'),
        _multiple_defs_notice_for('std:term:b'),
    )
    assert mult_defs_a not in app.warning.getvalue().lower()
    assert mult_defs_a not in app.status.getvalue().lower()
    assert mult_defs_b not in app.warning.getvalue().lower()
    assert mult_defs_b in app.status.getvalue().lower()


def _write_appconfig(dir: Path, language: str, prefix: str | None = None) -> Path:
    prefix = prefix or language
    (dir / prefix).mkdir(parents=True, exist_ok=True)
    (dir / prefix / 'conf.py').write_text(f'language = "{language}"', encoding='utf8')
    (dir / prefix / 'index.rst').write_text('index.rst', encoding='utf8')
    assert sorted(p.name for p in (dir / prefix).iterdir()) == ['conf.py', 'index.rst']
    assert (dir / prefix / 'index.rst').exists()
    return dir / prefix


def _build_inventory(srcdir: Path) -> Path:
    app = SphinxTestApp(srcdir=srcdir)
    app.build()
    sphinx.locale.translators.clear()
    return app.outdir / 'objects.inv'


def test_inventory_localization(tmp_path):
    # Build an app using Estonian (EE) locale
    srcdir_et = _write_appconfig(tmp_path, 'et')
    inventory_et = _build_inventory(srcdir_et)

    # Build the same app using English (US) locale
    srcdir_en = _write_appconfig(tmp_path, 'en')
    inventory_en = _build_inventory(srcdir_en)

    # Ensure that the inventory contents differ
    assert inventory_et.read_bytes() != inventory_en.read_bytes()



"""Test the other directives."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree(app):
    text = '.. toctree::\n\n   foo\n   bar/index\n   baz\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'foo'), (None, 'bar/index'), (None, 'baz')],
        includefiles=['foo', 'bar/index', 'baz'],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_relative_toctree(app):
    text = '.. toctree::\n\n   bar_1\n   bar_2\n   bar_3\n   ../quux\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'bar/index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[
            (None, 'bar/bar_1'),
            (None, 'bar/bar_2'),
            (None, 'bar/bar_3'),
            (None, 'quux'),
        ],
        includefiles=['bar/bar_1', 'bar/bar_2', 'bar/bar_3', 'quux'],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree_urls_and_titles(app):
    text = (
        '.. toctree::\n'
        '\n'
        '   Sphinx <https://www.sphinx-doc.org/>\n'
        '   https://readthedocs.org/\n'
        '   The BAR <bar/index>\n'
    )

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[
            ('Sphinx', 'https://www.sphinx-doc.org/'),
            (None, 'https://readthedocs.org/'),
            ('The BAR', 'bar/index'),
        ],
        includefiles=['bar/index'],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree_glob(app):
    text = '.. toctree::\n   :glob:\n\n   *\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'baz'), (None, 'foo'), (None, 'quux')],
        includefiles=['baz', 'foo', 'quux'],
    )

    # give both docname and glob (case1)
    text = '.. toctree::\n   :glob:\n\n   foo\n   *\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'foo'), (None, 'baz'), (None, 'quux')],
        includefiles=['foo', 'baz', 'quux'],
    )

    # give both docname and glob (case2)
    text = '.. toctree::\n   :glob:\n\n   *\n   foo\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'baz'), (None, 'foo'), (None, 'quux'), (None, 'foo')],
        includefiles=['baz', 'foo', 'quux', 'foo'],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree_glob_and_url(app):
    text = '.. toctree::\n   :glob:\n\n   https://example.com/?q=sphinx\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'https://example.com/?q=sphinx')],
        includefiles=[],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_reversed_toctree(app):
    text = '.. toctree::\n   :reversed:\n\n   foo\n   bar/index\n   baz\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'baz'), (None, 'bar/index'), (None, 'foo')],
        includefiles=['baz', 'bar/index', 'foo'],
    )


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree_class(app):
    text = '.. toctree::\n   :class: custom-toc\n\n   foo\n'
    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert doctree[0].attributes['classes'] == ['toctree-wrapper', 'custom-toc']


@pytest.mark.sphinx('html', testroot='toctree-glob')
def test_toctree_twice(app):
    text = '.. toctree::\n\n   foo\n   foo\n'

    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, [nodes.document, nodes.compound, addnodes.toctree])
    assert_node(
        doctree[0][0],
        entries=[(None, 'foo'), (None, 'foo')],
        includefiles=['foo', 'foo'],
    )


@pytest.mark.sphinx('html', testroot='directive-include')
def test_include_include_read_event(app: SphinxTestApp) -> None:
    sources_reported = []

    def source_read_handler(_app, relative_path, parent_docname, source):
        sources_reported.append((relative_path, parent_docname, source[0]))

    app.connect('include-read', source_read_handler)
    text = """\
.. include:: baz/baz.rst
   :start-line: 4
.. include:: text.txt
   :literal:
.. include:: bar.txt
"""
    app.env.find_files(app.config, app.builder)
    restructuredtext.parse(app, text, 'index')

    included_files = {filename.as_posix() for filename, p, s in sources_reported}
    # sources don't emit 'include-read'
    assert 'index.rst' not in included_files

    assert 'baz/baz.rst' in included_files

    # text was included as literal, no rst parsing
    assert 'text.txt' not in included_files

    # suffix not in source-suffixes
    assert 'bar.txt' in included_files

    assert (Path('baz/baz.rst'), 'index', '\nBaz was here.') in sources_reported


@pytest.mark.sphinx('html', testroot='directive-include')
def test_include_include_read_event_nested_includes(app):
    def source_read_handler(_app, _relative_path, _parent_docname, source):
        text = source[0].replace('#magical', 'amazing')
        source[0] = text

    app.connect('include-read', source_read_handler)
    text = '.. include:: baz/baz.rst\n'
    app.env.find_files(app.config, app.builder)
    doctree = restructuredtext.parse(app, text, 'index')
    assert_node(doctree, addnodes.document)
    assert len(doctree.children) == 3
    assert_node(doctree.children[1], nodes.paragraph)
    assert doctree.children[1].rawsource == 'The amazing foo.'



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import os
import subprocess
import xml.etree.ElementTree as ET
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING

import pytest

from sphinx.builders.epub3 import _XML_NAME_PATTERN
from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Self

    from sphinx.testing.util import SphinxTestApp


# check given command is runnable
def runnable(command: list[str]) -> bool:
    try:
        subprocess.run(command, capture_output=True, check=True)
        return True
    except (OSError, CalledProcessError):
        return False  # command not found or exit with non-zero


class EPUBElementTree:
    """Test helper for content.opf and toc.ncx"""

    namespaces = {
        'idpf': 'http://www.idpf.org/2007/opf',
        'dc': 'http://purl.org/dc/elements/1.1/',
        'ibooks': 'http://vocabulary.itunes.apple.com/rdf/ibooks/vocabulary-extensions-1.0/',
        'ncx': 'http://www.daisy.org/z3986/2005/ncx/',
        'xhtml': 'http://www.w3.org/1999/xhtml',
        'epub': 'http://www.idpf.org/2007/ops',
    }

    def __init__(self, tree: ET.Element) -> None:
        self.tree = tree

    @classmethod
    def fromstring(cls, string: str | bytes) -> Self:
        tree = ET.fromstring(string)  # NoQA: S314  # using known data in tests
        return cls(tree)

    def find(self, match: str) -> Self:
        ret = self.tree.find(match, namespaces=self.namespaces)
        assert ret is not None
        return self.__class__(ret)

    def findall(self, match: str) -> list[Self]:
        ret = self.tree.findall(match, namespaces=self.namespaces)
        return [self.__class__(e) for e in ret]

    @property
    def text(self) -> str | None:
        return self.tree.text

    @property
    def attrib(self) -> dict[str, str]:
        return self.tree.attrib

    def get(self, key: str) -> str | None:
        return self.tree.get(key)

    def __iter__(self) -> Iterator[Self]:
        for child in self.tree:
            yield self.__class__(child)


@pytest.mark.sphinx('epub', testroot='basic')
def test_build_epub(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / 'mimetype').read_text(
        encoding='utf8'
    ) == 'application/epub+zip'
    assert (app.outdir / 'META-INF' / 'container.xml').exists()

    # toc.ncx
    toc = EPUBElementTree.fromstring(
        (app.outdir / 'toc.ncx').read_text(encoding='utf8')
    )
    assert toc.find('./ncx:docTitle/ncx:text').text == 'Project name not set'

    # toc.ncx / head
    meta = list(toc.find('./ncx:head'))
    assert meta[0].attrib == {'name': 'dtb:uid', 'content': 'unknown'}
    assert meta[1].attrib == {'name': 'dtb:depth', 'content': '1'}
    assert meta[2].attrib == {'name': 'dtb:totalPageCount', 'content': '0'}
    assert meta[3].attrib == {'name': 'dtb:maxPageNumber', 'content': '0'}

    # toc.ncx / navMap
    navpoints = toc.findall('./ncx:navMap/ncx:navPoint')
    assert len(navpoints) == 1
    assert navpoints[0].attrib == {'id': 'navPoint1', 'playOrder': '1'}
    assert navpoints[0].find('./ncx:content').attrib == {'src': 'index.xhtml'}

    navlabel = navpoints[0].find('./ncx:navLabel/ncx:text')
    assert navlabel.text == 'The basic Sphinx documentation for testing'

    # content.opf
    opf = EPUBElementTree.fromstring(
        (app.outdir / 'content.opf').read_text(encoding='utf8')
    )

    # content.opf / metadata
    metadata = opf.find('./idpf:metadata')
    assert metadata.find('./dc:language').text == 'en'
    assert metadata.find('./dc:title').text == 'Project name not set'
    assert metadata.find('./dc:description').text == 'unknown'
    assert metadata.find('./dc:creator').text == 'Author name not set'
    assert metadata.find('./dc:contributor').text == 'unknown'
    assert metadata.find('./dc:publisher').text == 'Author name not set'
    assert metadata.find('./dc:rights').text is None
    assert metadata.find("./idpf:meta[@property='ibooks:version']").text is None
    assert (
        metadata.find("./idpf:meta[@property='ibooks:specified-fonts']").text == 'true'
    )
    assert metadata.find("./idpf:meta[@property='ibooks:binding']").text == 'true'
    assert (
        metadata.find("./idpf:meta[@property='ibooks:scroll-axis']").text == 'vertical'
    )

    # content.opf / manifest
    manifest = opf.find('./idpf:manifest')
    items = list(manifest)
    assert items[0].attrib == {
        'id': 'ncx',
        'href': 'toc.ncx',
        'media-type': 'application/x-dtbncx+xml',
    }
    assert items[1].attrib == {
        'id': 'nav',
        'href': 'nav.xhtml',
        'media-type': 'application/xhtml+xml',
        'properties': 'nav',
    }
    assert items[2].attrib == {
        'id': 'epub-0',
        'href': 'genindex.xhtml',
        'media-type': 'application/xhtml+xml',
    }
    assert items[3].attrib == {
        'id': 'epub-1',
        'href': 'index.xhtml',
        'media-type': 'application/xhtml+xml',
    }

    for i, item in enumerate(items[2:]):
        # items are named as epub-NN
        assert item.get('id') == 'epub-%d' % i

    # content.opf / spine
    spine = opf.find('./idpf:spine')
    itemrefs = list(spine)
    assert spine.get('toc') == 'ncx'
    assert spine.get('page-progression-direction') == 'ltr'
    assert itemrefs[0].get('idref') == 'epub-1'
    assert itemrefs[1].get('idref') == 'epub-0'

    # content.opf / guide
    reference = opf.find('./idpf:guide/idpf:reference')
    assert reference.get('type') == 'toc'
    assert reference.get('title') == 'Table of Contents'
    assert reference.get('href') == 'index.xhtml'

    # nav.xhtml
    nav = EPUBElementTree.fromstring(
        (app.outdir / 'nav.xhtml').read_text(encoding='utf8')
    )
    assert nav.attrib == {
        'lang': 'en',
        '{http://www.w3.org/XML/1998/namespace}lang': 'en',
    }
    assert nav.find('./xhtml:head/xhtml:title').text == 'Table of Contents'

    # nav.xhtml / nav
    navlist = nav.find('./xhtml:body/xhtml:nav')
    tocs = navlist.findall('./xhtml:ol/xhtml:li')
    assert navlist.find('./xhtml:h1').text == 'Table of Contents'
    assert len(tocs) == 1
    assert tocs[0].find('./xhtml:a').get('href') == 'index.xhtml'
    assert (
        tocs[0].find('./xhtml:a').text == 'The basic Sphinx documentation for testing'
    )


@pytest.mark.sphinx(
    'epub',
    testroot='footnotes',
    confoverrides={'epub_cover': ('_images/rimg.png', None)},
)
def test_epub_cover(app: SphinxTestApp) -> None:
    app.build()

    # content.opf / metadata
    opf = EPUBElementTree.fromstring(
        (app.outdir / 'content.opf').read_text(encoding='utf8')
    )
    cover_image = opf.find(
        "./idpf:manifest/idpf:item[@href='%s']" % app.config.epub_cover[0]
    )
    cover = opf.find("./idpf:metadata/idpf:meta[@name='cover']")
    assert cover
    assert cover.get('content') == cover_image.get('id')


@pytest.mark.sphinx('epub', testroot='toctree')
def test_nested_toc(app: SphinxTestApp) -> None:
    app.build()

    # toc.ncx
    toc = EPUBElementTree.fromstring((app.outdir / 'toc.ncx').read_bytes())
    assert toc.find('./ncx:docTitle/ncx:text').text == 'Project name not set'

    # toc.ncx / navPoint
    def toc_navpoint_navinfo(
        elem: EPUBElementTree,
    ) -> tuple[str | None, str | None, str | None, str | None]:
        label = elem.find('./ncx:navLabel/ncx:text')
        content = elem.find('./ncx:content')
        return elem.get('id'), elem.get('playOrder'), content.get('src'), label.text

    navpoints = toc.findall('./ncx:navMap/ncx:navPoint')
    assert len(navpoints) == 4
    assert toc_navpoint_navinfo(navpoints[0]) == (
        'navPoint1',
        '1',
        'index.xhtml',
        'Welcome to Sphinx Tests’s documentation!',
    )
    assert navpoints[0].findall('./ncx:navPoint') == []

    # toc.ncx / nested navPoints
    assert toc_navpoint_navinfo(navpoints[1]) == ('navPoint2', '2', 'foo.xhtml', 'foo')
    navchildren = navpoints[1].findall('./ncx:navPoint')
    assert len(navchildren) == 4
    assert toc_navpoint_navinfo(navchildren[0]) == (
        'navPoint3',
        '2',
        'foo.xhtml',
        'foo',
    )
    assert toc_navpoint_navinfo(navchildren[1]) == (
        'navPoint4',
        '3',
        'quux.xhtml',
        'quux',
    )
    assert toc_navpoint_navinfo(navchildren[2]) == (
        'navPoint5',
        '4',
        'foo.xhtml#foo-1',
        'foo.1',
    )
    assert toc_navpoint_navinfo(navchildren[3]) == (
        'navPoint8',
        '6',
        'foo.xhtml#foo-2',
        'foo.2',
    )

    # nav.xhtml / nav
    def nav_nav_navinfo(elem: EPUBElementTree) -> tuple[str | None, str | None]:
        anchor = elem.find('./xhtml:a')
        return anchor.get('href'), anchor.text

    nav = EPUBElementTree.fromstring((app.outdir / 'nav.xhtml').read_bytes())
    tocs = nav.findall('./xhtml:body/xhtml:nav/xhtml:ol/xhtml:li')
    assert len(tocs) == 4
    assert nav_nav_navinfo(tocs[0]) == (
        'index.xhtml',
        'Welcome to Sphinx Tests’s documentation!',
    )
    assert tocs[0].findall('./xhtml:ol') == []

    # nav.xhtml / nested toc
    assert nav_nav_navinfo(tocs[1]) == ('foo.xhtml', 'foo')
    tocchildren = tocs[1].findall('./xhtml:ol/xhtml:li')
    assert len(tocchildren) == 3
    assert nav_nav_navinfo(tocchildren[0]) == ('quux.xhtml', 'quux')
    assert nav_nav_navinfo(tocchildren[1]) == ('foo.xhtml#foo-1', 'foo.1')
    assert nav_nav_navinfo(tocchildren[2]) == ('foo.xhtml#foo-2', 'foo.2')

    grandchild = tocchildren[1].findall('./xhtml:ol/xhtml:li')
    assert len(grandchild) == 1
    assert nav_nav_navinfo(grandchild[0]) == ('foo.xhtml#foo-1-1', 'foo.1-1')


@pytest.mark.sphinx('epub', testroot='need-escaped')
def test_escaped_toc(app: SphinxTestApp) -> None:
    app.build()

    # toc.ncx
    toc = EPUBElementTree.fromstring((app.outdir / 'toc.ncx').read_bytes())
    assert toc.find('./ncx:docTitle/ncx:text').text == 'need <b>"escaped"</b> project'

    # toc.ncx / navPoint
    def navpoint_navinfo(
        elem: EPUBElementTree,
    ) -> tuple[str | None, str | None, str | None, str | None]:
        label = elem.find('./ncx:navLabel/ncx:text')
        content = elem.find('./ncx:content')
        ret = elem.get('id'), elem.get('playOrder'), content.get('src'), label.text
        return ret

    navpoints = toc.findall('./ncx:navMap/ncx:navPoint')
    assert len(navpoints) == 4
    assert navpoint_navinfo(navpoints[0]) == (
        'navPoint1',
        '1',
        'index.xhtml',
        "Welcome to Sphinx Tests's documentation!",
    )
    assert navpoints[0].findall('./ncx:navPoint') == []

    # toc.ncx / nested navPoints
    assert navpoint_navinfo(navpoints[1]) == ('navPoint2', '2', 'foo.xhtml', '<foo>')
    navchildren = navpoints[1].findall('./ncx:navPoint')
    assert len(navchildren) == 4
    assert navpoint_navinfo(navchildren[0]) == ('navPoint3', '2', 'foo.xhtml', '<foo>')
    assert navpoint_navinfo(navchildren[1]) == ('navPoint4', '3', 'quux.xhtml', 'quux')
    assert navpoint_navinfo(navchildren[2]) == (
        'navPoint5',
        '4',
        'foo.xhtml#foo-1',
        'foo “1”',
    )
    assert navpoint_navinfo(navchildren[3]) == (
        'navPoint8',
        '6',
        'foo.xhtml#foo-2',
        'foo.2',
    )

    # nav.xhtml / nav
    def nav_navinfo(elem: EPUBElementTree) -> tuple[str | None, str | None]:
        anchor = elem.find('./xhtml:a')
        return anchor.get('href'), anchor.text

    nav = EPUBElementTree.fromstring((app.outdir / 'nav.xhtml').read_bytes())
    tocs = nav.findall('./xhtml:body/xhtml:nav/xhtml:ol/xhtml:li')
    assert len(tocs) == 4
    assert nav_navinfo(tocs[0]) == (
        'index.xhtml',
        "Welcome to Sphinx Tests's documentation!",
    )
    assert tocs[0].findall('./xhtml:ol') == []

    # nav.xhtml / nested toc
    assert nav_navinfo(tocs[1]) == ('foo.xhtml', '<foo>')
    tocchildren = tocs[1].findall('./xhtml:ol/xhtml:li')
    assert len(tocchildren) == 3
    assert nav_navinfo(tocchildren[0]) == ('quux.xhtml', 'quux')
    assert nav_navinfo(tocchildren[1]) == ('foo.xhtml#foo-1', 'foo “1”')
    assert nav_navinfo(tocchildren[2]) == ('foo.xhtml#foo-2', 'foo.2')

    grandchild = tocchildren[1].findall('./xhtml:ol/xhtml:li')
    assert len(grandchild) == 1
    assert nav_navinfo(grandchild[0]) == ('foo.xhtml#foo-1-1', 'foo.1-1')


@pytest.mark.sphinx('epub', testroot='basic')
def test_epub_writing_mode(app: SphinxTestApp) -> None:
    # horizontal (default)
    app.build(force_all=True)

    # horizontal / page-progression-direction
    opf = EPUBElementTree.fromstring(
        (app.outdir / 'content.opf').read_text(encoding='utf8')
    )
    assert opf.find('./idpf:spine').get('page-progression-direction') == 'ltr'

    # horizontal / ibooks:scroll-axis
    metadata = opf.find('./idpf:metadata')
    assert (
        metadata.find("./idpf:meta[@property='ibooks:scroll-axis']").text == 'vertical'
    )

    # horizontal / writing-mode (CSS)
    css = (app.outdir / '_static' / 'epub.css').read_text(encoding='utf8')
    assert 'writing-mode: horizontal-tb;' in css

    # vertical
    app.config.epub_writing_mode = 'vertical'
    (app.outdir / 'index.xhtml').unlink()  # forcely rebuild
    app.build()

    # vertical / page-progression-direction
    opf = EPUBElementTree.fromstring(
        (app.outdir / 'content.opf').read_text(encoding='utf8')
    )
    assert opf.find('./idpf:spine').get('page-progression-direction') == 'rtl'

    # vertical / ibooks:scroll-axis
    metadata = opf.find('./idpf:metadata')
    assert (
        metadata.find("./idpf:meta[@property='ibooks:scroll-axis']").text
        == 'horizontal'
    )

    # vertical / writing-mode (CSS)
    css = (app.outdir / '_static' / 'epub.css').read_text(encoding='utf8')
    assert 'writing-mode: vertical-rl;' in css


@pytest.mark.sphinx('epub', testroot='epub-anchor-id')
def test_epub_anchor_id(app: SphinxTestApp) -> None:
    app.build()

    html = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert '<p id="std-setting-STATICFILES_FINDERS">blah blah blah</p>' in html
    assert (
        '<span id="std-setting-STATICFILES_SECTION"></span><h1>blah blah blah</h1>'
    ) in html
    assert (
        'see <a class="reference internal" href="#std-setting-STATICFILES_FINDERS">'
    ) in html


@pytest.mark.sphinx('epub', testroot='html_assets')
def test_epub_assets(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    # epub_sytlesheets (same as html_css_files)
    content = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/css/style.css" />'
    ) in content
    assert (
        '<link media="print" rel="stylesheet" title="title" type="text/css" '
        'href="https://example.com/custom.css" />'
    ) in content


@pytest.mark.sphinx(
    'epub',
    testroot='html_assets',
    confoverrides={'epub_css_files': ['css/epub.css']},
)
def test_epub_css_files(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    # epub_css_files
    content = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/css/epub.css" />'
    ) in content

    # files in html_css_files are not outputted
    assert (
        '<link rel="stylesheet" type="text/css" href="_static/css/style.css" />'
    ) not in content
    assert (
        '<link media="print" rel="stylesheet" title="title" type="text/css" '
        'href="https://example.com/custom.css" />'
    ) not in content


@pytest.mark.sphinx('epub', testroot='roles-download')
def test_html_download_role(app: SphinxTestApp) -> None:
    app.build()
    assert not (app.outdir / '_downloads' / 'dummy.dat').exists()

    content = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert (
        '<li><p><code class="xref download docutils literal notranslate">'
        '<span class="pre">dummy.dat</span></code></p></li>'
    ) in content
    assert (
        '<li><p><code class="xref download docutils literal notranslate">'
        '<span class="pre">not_found.dat</span></code></p></li>'
    ) in content
    assert (
        '<li><p><code class="xref download docutils literal notranslate">'
        '<span class="pre">Sphinx</span> <span class="pre">logo</span></code>'
        '<span class="link-target"> [https://www.sphinx-doc.org/en/master'
        '/_static/sphinx-logo.svg]</span></p></li>'
    ) in content


@pytest.mark.sphinx('epub', testroot='toctree-duplicated')
def test_duplicated_toctree_entry(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert 'WARNING: duplicated ToC entry found: foo.xhtml' in app.warning.getvalue()


@pytest.mark.skipif(
    'DO_EPUBCHECK' not in os.environ,
    reason='Skipped because DO_EPUBCHECK is not set',
)
@pytest.mark.sphinx('epub', testroot='root')
def test_run_epubcheck(app: SphinxTestApp) -> None:
    app.build()

    if not runnable(['java', '-version']):
        pytest.skip('Unable to run Java; skipping test')

    epubcheck = Path(os.environ.get('EPUBCHECK_PATH', '/usr/share/java/epubcheck.jar'))
    if not epubcheck.exists():
        pytest.skip('Could not find epubcheck; skipping test')

    try:
        subprocess.run(
            ['java', '-jar', epubcheck, app.outdir / 'SphinxTests.epub'],  # NoQA: S607
            capture_output=True,
            check=True,
        )
    except CalledProcessError as exc:
        print(exc.stdout.decode('utf-8'))
        print(exc.stderr.decode('utf-8'))
        msg = f'epubcheck exited with return code {exc.returncode}'
        raise AssertionError(msg) from exc


def test_xml_name_pattern_check() -> None:
    assert _XML_NAME_PATTERN.match('id-pub')
    assert _XML_NAME_PATTERN.match('webpage')
    assert not _XML_NAME_PATTERN.match('1bfda21')


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx('epub', testroot='images')
def test_copy_images(app: SphinxTestApp) -> None:
    app.build()

    images_dir = Path(app.outdir) / '_images'
    images = {image.name for image in images_dir.rglob('*')}
    images.discard('python-logo.png')
    assert images == {
        # 'ba30773957c3fe046897111afd65a80b81cad089.png',  # epub: image from data:image/png URI in source
        'img.png',
        'rimg.png',
        'rimg1.png',
        'svgimg.svg',
        'testimäge.png',
    }



"""Test the intersphinx extension."""

from __future__ import annotations

import posixpath
import re
import zlib
from http.server import BaseHTTPRequestHandler
from io import BytesIO
from typing import TYPE_CHECKING

from sphinx.ext.intersphinx._shared import InventoryAdapter
from sphinx.testing.util import SphinxTestApp
from sphinx.util.inventory import _InventoryItem

from tests.utils import http_server

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import BinaryIO


BASE_CONFIG = {
    'extensions': ['sphinx.ext.intersphinx'],
    'intersphinx_timeout': 0.1,
}


class InventoryEntry:
    """Entry in the Intersphinx inventory."""

    __slots__ = (
        'name',
        'display_name',
        'domain_name',
        'object_type',
        'uri',
        'anchor',
        'priority',
    )

    def __init__(
        self,
        name: str = 'this',
        *,
        display_name: str | None = None,
        domain_name: str = 'py',
        object_type: str = 'obj',
        uri: str = 'index.html',
        anchor: str = '',
        priority: int = 0,
    ):
        if anchor.endswith(name):
            anchor = anchor.removesuffix(name) + '$'

        if anchor:
            uri += '#' + anchor

        if display_name is None or display_name == name:
            display_name = '-'

        self.name = name
        self.display_name = display_name
        self.domain_name = domain_name
        self.object_type = object_type
        self.uri = uri
        self.anchor = anchor
        self.priority = priority

    def format(self) -> str:
        """Format the entry as it appears in the inventory file."""
        return (
            f'{self.name} {self.domain_name}:{self.object_type} '
            f'{self.priority} {self.uri} {self.display_name}\n'
        )


class IntersphinxProject:
    def __init__(
        self,
        *,
        name: str = 'spam',
        version: str | int = 1,
        baseurl: str = '',
        baseuri: str = '',
        file: str | None = None,
    ) -> None:
        #: The project name.
        self.name = name
        #: The escaped project name.
        self.safe_name = re.sub(r'\\s+', ' ', name)

        #: The project version as a string.
        self.version = version = str(version)
        #: The escaped project version.
        self.safe_version = re.sub(r'\\s+', ' ', version)

        #: The project base URL (e.g., http://localhost:9341).
        self.baseurl = baseurl
        #: The project base URI, relative to *baseurl* (e.g., 'spam').
        self.uri = baseuri
        #: The project URL, as specified in :confval:`intersphinx_mapping`.
        self.url = posixpath.join(baseurl, baseuri)
        #: The project local file, if any.
        self.file = file

    @property
    def record(self) -> dict[str, tuple[str | None, str | None]]:
        """The :confval:`intersphinx_mapping` record for this project."""
        return {self.name: (self.url, self.file)}

    def normalise(self, entry: InventoryEntry) -> tuple[str, _InventoryItem]:
        """Format an inventory entry as if it were part of this project."""
        return entry.name, _InventoryItem(
            project_name=self.safe_name,
            project_version=self.safe_version,
            uri=posixpath.join(self.url, entry.uri),
            display_name=entry.display_name,
        )


class FakeInventory:
    protocol_version: int

    def __init__(self, project: IntersphinxProject | None = None) -> None:
        self.project = project or IntersphinxProject()

    def serialise(self, entries: Iterable[InventoryEntry] | None = None) -> bytes:
        buffer = BytesIO()
        self._write_headers(buffer)
        entries = entries or [InventoryEntry()]
        self._write_body(buffer, (item.format().encode() for item in entries))
        return buffer.getvalue()

    def _write_headers(self, buffer: BinaryIO) -> None:
        headers = (
            f'# Sphinx inventory version {self.protocol_version}\n'
            f'# Project: {self.project.safe_name}\n'
            f'# Version: {self.project.safe_version}\n'
        ).encode()
        buffer.write(headers)

    def _write_body(self, buffer: BinaryIO, lines: Iterable[bytes]) -> None:
        raise NotImplementedError


class FakeInventoryV2(FakeInventory):
    protocol_version = 2

    def _write_headers(self, buffer: BinaryIO) -> None:
        super()._write_headers(buffer)
        buffer.write(b'# The remainder of this file is compressed using zlib.\n')

    def _write_body(self, buffer: BinaryIO, lines: Iterable[bytes]) -> None:
        compressor = zlib.compressobj(9)
        buffer.writelines(map(compressor.compress, lines))
        buffer.write(compressor.flush())


class SingleEntryProject(IntersphinxProject):
    name = 'spam'
    port = 9341  # needed since otherwise it's an automatic port

    def __init__(
        self,
        version: int,
        route: str,
        *,
        item_name: str = 'ham',
        domain_name: str = 'py',
        object_type: str = 'module',
    ) -> None:
        super().__init__(
            name=self.name,
            version=version,
            baseurl=f'http://localhost:{self.port}',
            baseuri=route,
        )
        self.item_name = item_name
        self.domain_name = domain_name
        self.object_type = object_type
        self.reftype = f'{domain_name}:{object_type}'

    def make_entry(self) -> InventoryEntry:
        """Get an inventory entry for this project."""
        name = f'{self.item_name}_{self.version}'
        return InventoryEntry(
            name, domain_name=self.domain_name, object_type=self.object_type
        )


def make_inventory_handler(
    *projects: SingleEntryProject,
) -> type[BaseHTTPRequestHandler]:
    name, port = projects[0].name, projects[0].port
    assert all(p.name == name for p in projects)
    assert all(p.port == port for p in projects)

    class InventoryHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200, 'OK')

            data = b''
            for project in projects:
                # create the data to return depending on the endpoint
                if self.path.startswith(f'/{project.uri}/'):
                    entry = project.make_entry()
                    data = FakeInventoryV2(project).serialise([entry])
                    break

            self.send_header('Content-Length', str(len(data)))
            self.end_headers()
            self.wfile.write(data)

        def log_message(*args, **kwargs):
            pass

    return InventoryHandler


def test_intersphinx_project_fixture() -> None:
    # check that our fixture class is correct
    project = SingleEntryProject(1, 'route')
    assert project.url == 'http://localhost:9341/route'


def test_load_mappings_cache(tmp_path):
    tmp_path.joinpath('conf.py').touch()
    tmp_path.joinpath('index.rst').touch()
    project = SingleEntryProject(1, 'a')

    InventoryHandler = make_inventory_handler(project)
    with http_server(InventoryHandler, port=project.port):
        # clean build
        confoverrides = BASE_CONFIG | {'intersphinx_mapping': project.record}
        app = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides)
        app.build()
        app.cleanup()

    # the inventory when querying the 'old' URL
    entry = project.make_entry()
    item = dict((project.normalise(entry),))
    inventories = InventoryAdapter(app.env)
    assert list(inventories.cache) == ['http://localhost:9341/a']
    e_name, e_time, e_inv = inventories.cache['http://localhost:9341/a']
    assert e_name == 'spam'
    assert e_inv == {'py:module': item}
    assert inventories.named_inventory == {'spam': {'py:module': item}}


def test_load_mappings_cache_update(tmp_path):
    tmp_path.joinpath('conf.py').touch()
    tmp_path.joinpath('index.rst').touch()
    old_project = SingleEntryProject(1337, 'old')
    new_project = SingleEntryProject(1701, 'new')

    InventoryHandler = make_inventory_handler(old_project, new_project)
    with http_server(InventoryHandler, port=SingleEntryProject.port):
        # build normally to create an initial cache
        confoverrides1 = BASE_CONFIG | {'intersphinx_mapping': old_project.record}
        app1 = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides1)
        app1.build()
        app1.cleanup()

        # switch to new url and assert that the old URL is no more stored
        confoverrides2 = BASE_CONFIG | {'intersphinx_mapping': new_project.record}
        app2 = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides2)
        app2.build()
        app2.cleanup()

    entry = new_project.make_entry()
    item = dict((new_project.normalise(entry),))
    inventories = InventoryAdapter(app2.env)
    # check that the URLs were changed accordingly
    assert list(inventories.cache) == ['http://localhost:9341/new']
    e_name, e_time, e_inv = inventories.cache['http://localhost:9341/new']
    assert e_name == 'spam'
    assert e_inv == {'py:module': item}
    assert inventories.named_inventory == {'spam': {'py:module': item}}


def test_load_mappings_cache_revert_update(tmp_path):
    tmp_path.joinpath('conf.py').touch()
    tmp_path.joinpath('index.rst').touch()
    old_project = SingleEntryProject(1337, 'old')
    new_project = SingleEntryProject(1701, 'new')

    InventoryHandler = make_inventory_handler(old_project, new_project)
    with http_server(InventoryHandler, port=SingleEntryProject.port):
        # build normally to create an initial cache
        confoverrides1 = BASE_CONFIG | {'intersphinx_mapping': old_project.record}
        app1 = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides1)
        app1.build()
        app1.cleanup()

        # switch to new url and build
        confoverrides2 = BASE_CONFIG | {'intersphinx_mapping': new_project.record}
        app2 = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides2)
        app2.build()
        app2.cleanup()

        # switch back to old url (reuse 'old_item')
        confoverrides3 = BASE_CONFIG | {'intersphinx_mapping': old_project.record}
        app3 = SphinxTestApp('dummy', srcdir=tmp_path, confoverrides=confoverrides3)
        app3.build()
        app3.cleanup()

    entry = old_project.make_entry()
    item = dict((old_project.normalise(entry),))
    inventories = InventoryAdapter(app3.env)
    # check that the URLs were changed accordingly
    assert list(inventories.cache) == ['http://localhost:9341/old']
    e_name, e_time, e_inv = inventories.cache['http://localhost:9341/old']
    assert e_name == 'spam'
    assert e_inv == {'py:module': item}
    assert inventories.named_inventory == {'spam': {'py:module': item}}



from __future__ import annotations

import time
from typing import TYPE_CHECKING

import pytest

from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Iterator

    from sphinx.testing.util import SphinxTestApp

LT = time.localtime()
LT_NEW = (2009, *LT[1:], LT.tm_zone, LT.tm_gmtoff)
LOCALTIME_2009 = type(LT)(LT_NEW)


@pytest.fixture
def no_source_date_year(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    """Explicitly clear SOURCE_DATE_EPOCH from the environment; this
    fixture can be used to ensure that copyright substitution logic
    does not occur during selected test cases.
    """
    with monkeypatch.context() as m:
        m.delenv('SOURCE_DATE_EPOCH', raising=False)
        yield


@pytest.fixture(
    params=[
        1199145600,  # 2008-01-01 00:00:00
        1199145599,  # 2007-12-31 23:59:59
    ]
)
def source_date_year(
    request: pytest.FixtureRequest,
    monkeypatch: pytest.MonkeyPatch,
) -> Iterator[int]:
    source_date_epoch = request.param
    with monkeypatch.context() as m:
        m.setattr(time, 'localtime', lambda *a: LOCALTIME_2009)
        m.setenv('SOURCE_DATE_EPOCH', str(source_date_epoch))
        yield time.gmtime(source_date_epoch).tm_year


@pytest.mark.sphinx('html', testroot='copyright-multiline')
def test_html_multi_line_copyright(
    no_source_date_year: None, app: SphinxTestApp
) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'index.html').read_text(encoding='utf-8')

    # check the copyright footer line by line (empty lines ignored)
    assert '  &#169; Copyright 2006.<br/>\n' in content
    assert '  &#169; Copyright 2006-2009, Alice.<br/>\n' in content
    assert '  &#169; Copyright 2010-2013, Bob.<br/>\n' in content
    assert '  &#169; Copyright 2014-2017, Charlie.<br/>\n' in content
    assert '  &#169; Copyright 2018-2021, David.<br/>\n' in content
    assert '  &#169; Copyright 2022-2025, Eve.' in content

    # check the raw copyright footer block (empty lines included)
    assert (
        '      &#169; Copyright 2006.<br/>\n'
        '    \n'
        '      &#169; Copyright 2006-2009, Alice.<br/>\n'
        '    \n'
        '      &#169; Copyright 2010-2013, Bob.<br/>\n'
        '    \n'
        '      &#169; Copyright 2014-2017, Charlie.<br/>\n'
        '    \n'
        '      &#169; Copyright 2018-2021, David.<br/>\n'
        '    \n'
        '      &#169; Copyright 2022-2025, Eve.'
    ) in content


@pytest.mark.sphinx('html', testroot='copyright-multiline')
def test_html_multi_line_copyright_sde(
    source_date_year: None, app: SphinxTestApp
) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'index.html').read_text(encoding='utf-8')

    # check the copyright footer line by line (empty lines ignored)
    assert '  &#169; Copyright 2006.<br/>\n' in content
    assert f'  &#169; Copyright 2006-{source_date_year}, Alice.<br/>\n' in content
    assert '  &#169; Copyright 2010-2013, Bob.<br/>\n' in content
    assert '  &#169; Copyright 2014-2017, Charlie.<br/>\n' in content
    assert '  &#169; Copyright 2018-2021, David.<br/>\n' in content
    assert '  &#169; Copyright 2022-2025, Eve.' in content

    # check the raw copyright footer block (empty lines included)
    assert (
        '      &#169; Copyright 2006.<br/>\n'
        '    \n'
        f'      &#169; Copyright 2006-{source_date_year}, Alice.<br/>\n'
        '    \n'
        '      &#169; Copyright 2010-2013, Bob.<br/>\n'
        '    \n'
        '      &#169; Copyright 2014-2017, Charlie.<br/>\n'
        '    \n'
        '      &#169; Copyright 2018-2021, David.<br/>\n'
        '    \n'
        '      &#169; Copyright 2022-2025, Eve.'
    ) in content



from __future__ import annotations

from datetime import datetime
from typing import Any  # NoQA: TC003

CONSTANT = 'foo'
SENTINEL = object()


def foo(
    name: str = CONSTANT,
    sentinel: Any = SENTINEL,
    now: datetime = datetime.now(),  # NoQA: B008,DTZ005
    color: int = 0xFFFFFF,
    *,
    kwarg1,
    kwarg2=0xFFFFFF,
) -> None:
    """docstring"""


class Class:
    """docstring"""

    def meth(
        self,
        name: str = CONSTANT,
        sentinel: Any = SENTINEL,
        now: datetime = datetime.now(),  # NoQA: B008,DTZ005
        color: int = 0xFFFFFF,
        *,
        kwarg1,
        kwarg2=0xFFFFFF,
    ) -> None:
        """docstring"""

    @classmethod
    def clsmeth(
        cls,
        name: str = CONSTANT,
        sentinel: Any = SENTINEL,
        now: datetime = datetime.now(),  # NoQA: B008,DTZ005
        color: int = 0xFFFFFF,
        *,
        kwarg1,
        kwarg2=0xFFFFFF,
    ) -> None:
        """docstring"""


get_sentinel = lambda custom=SENTINEL: custom  # NoQA: E731
"""docstring"""


class MultiLine:
    """docstring"""

    # The properties will raise a silent SyntaxError because "lambda self: 1"
    # will be detected as a function to update the default values of. However,
    # only prop3 will not fail because it's on a single line whereas the others
    # will fail to parse.

    # fmt: off
    prop1 = property(
      lambda self: 1, doc='docstring')

    prop2 = property(
      lambda self: 2, doc='docstring'
    )

    prop3 = property(lambda self: 3, doc='docstring')

    prop4 = (property
    (lambda self: 4, doc='docstring'))

    prop5 = property\
    (lambda self: 5, doc='docstring')  # NoQA: E211
    # fmt: on



"""Test dirhtml builder."""

from __future__ import annotations

import posixpath
from typing import TYPE_CHECKING

import pytest

from sphinx.util.inventory import InventoryFile, _InventoryItem

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('dirhtml', testroot='builder-dirhtml')
def test_dirhtml(app: SphinxTestApp) -> None:
    app.build()

    assert (app.outdir / 'index.html').exists()
    assert (app.outdir / 'foo/index.html').exists()
    assert (app.outdir / 'foo/foo_1/index.html').exists()
    assert (app.outdir / 'foo/foo_2/index.html').exists()
    assert (app.outdir / 'bar/index.html').exists()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert 'href="foo/"' in content
    assert 'href="foo/foo_1/"' in content
    assert 'href="foo/foo_2/"' in content
    assert 'href="bar/"' in content

    # objects.inv
    # See: https://github.com/sphinx-doc/sphinx/issues/7095
    with (app.outdir / 'objects.inv').open('rb') as f:
        invdata = InventoryFile.load(f, 'path/to', posixpath.join)

    assert 'index' in invdata.get('std:doc', {})
    assert invdata['std:doc']['index'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='path/to/',
        display_name='-',
    )

    assert 'foo/index' in invdata.get('std:doc', {})
    assert invdata['std:doc']['foo/index'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='path/to/foo/',
        display_name='-',
    )

    assert 'index' in invdata.get('std:label', {})
    assert invdata['std:label']['index'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='path/to/#index',
        display_name='-',
    )

    assert 'foo' in invdata.get('std:label', {})
    assert invdata['std:label']['foo'] == _InventoryItem(
        project_name='Project name not set',
        project_version='',
        uri='path/to/foo/#foo',
        display_name='foo/index',
    )



"""Tests util functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.util import logging
from sphinx.util.display import (
    SkipProgressMessage,
    display_chunk,
    progress_message,
    status_iterator,
)

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


def test_display_chunk() -> None:
    assert display_chunk('hello') == 'hello'
    assert display_chunk(['hello']) == 'hello'
    assert display_chunk(['hello', 'sphinx', 'world']) == 'hello .. world'
    assert display_chunk(('hello',)) == 'hello'
    assert display_chunk(('hello', 'sphinx', 'world')) == 'hello .. world'


@pytest.mark.sphinx('dummy', testroot='root')
def test_status_iterator_length_0(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)

    # test for status_iterator (length=0)
    app.status.seek(0)
    app.status.truncate(0)
    yields = list(status_iterator(['hello', 'sphinx', 'world'], 'testing ... '))
    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing ... hello sphinx world \n' in output
    assert yields == ['hello', 'sphinx', 'world']


@pytest.mark.sphinx('dummy', testroot='root')
def test_status_iterator_verbosity_0(app, monkeypatch):
    monkeypatch.setenv('FORCE_COLOR', '1')
    logging.setup(app, app.status, app.warning)

    # test for status_iterator (verbosity=0)
    app.status.seek(0)
    app.status.truncate(0)
    yields = status_iterator(
        ['hello', 'sphinx', 'world'], 'testing ... ', length=3, verbosity=0
    )
    assert list(yields) == ['hello', 'sphinx', 'world']
    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing ... [ 33%] hello\r' in output
    assert 'testing ... [ 67%] sphinx\r' in output
    assert 'testing ... [100%] world\r\n' in output


@pytest.mark.sphinx('dummy', testroot='root')
def test_status_iterator_verbosity_1(app, monkeypatch):
    monkeypatch.setenv('FORCE_COLOR', '1')
    logging.setup(app, app.status, app.warning)

    # test for status_iterator (verbosity=1)
    app.status.seek(0)
    app.status.truncate(0)
    yields = status_iterator(
        ['hello', 'sphinx', 'world'], 'testing ... ', length=3, verbosity=1
    )
    assert list(yields) == ['hello', 'sphinx', 'world']
    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing ... [ 33%] hello\n' in output
    assert 'testing ... [ 67%] sphinx\n' in output
    assert 'testing ... [100%] world\n\n' in output


@pytest.mark.sphinx('html', testroot='root')
def test_progress_message(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    # standard case
    with progress_message('testing'):
        logger.info('blah ', nonl=True)

    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing... blah done\n' in output

    # skipping case
    with progress_message('testing'):
        raise SkipProgressMessage('Reason: %s', 'error')  # NoQA: EM101,TRY003

    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing... skipped\nReason: error\n' in output

    # error case
    try:
        with progress_message('testing'):
            raise RuntimeError  # NoQA: TRY301
    except Exception:
        pass

    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing... failed\n' in output

    # decorator
    @progress_message('testing')
    def func():
        logger.info('in func ', nonl=True)

    func()
    output = strip_escape_sequences(app.status.getvalue())
    assert 'testing... in func done\n' in output



"""Tests the Python Domain"""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_sig_keyword,
    desc_sig_punctuation,
    desc_sig_space,
    desc_signature,
    pending_xref,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list(app):
    text = (
        '.. py:module:: example\n'
        '.. py:class:: Class\n'
        '\n'
        '   :meta blah: this meta-field must not show up in the toc-tree\n'
        '   :param str name: blah blah\n'
        '   :meta another meta field:\n'
        '   :param age: blah blah\n'
        '   :type age: int\n'
        '   :param items: blah blah\n'
        '   :type items: Tuple[str, ...]\n'
        '   :param Dict[str, str] params: blah blah\n'
    )
    doctree = restructuredtext.parse(app, text)
    print(doctree)

    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, nodes.field_list, nodes.field],
                ),
            ],
        ),
    )
    assert_node(
        doctree[3][1][0][0],
        (
            [nodes.field_name, 'Parameters'],
            [
                nodes.field_body,
                nodes.bullet_list,
                (
                    [nodes.list_item, nodes.paragraph],
                    [nodes.list_item, nodes.paragraph],
                    [nodes.list_item, nodes.paragraph],
                    [nodes.list_item, nodes.paragraph],
                ),
            ],
        ),
    )

    # :param str name:
    assert_node(
        doctree[3][1][0][0][1][0][0][0],
        (
            [addnodes.literal_strong, 'name'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'str'],
            ')',
            ' -- ',
            'blah blah',
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][0][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='str',
        **{'py:module': 'example', 'py:class': 'Class'},
    )

    # :param age: + :type age:
    assert_node(
        doctree[3][1][0][0][1][0][1][0],
        (
            [addnodes.literal_strong, 'age'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'int'],
            ')',
            ' -- ',
            'blah blah',
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][1][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='int',
        **{'py:module': 'example', 'py:class': 'Class'},
    )

    # :param items: + :type items:
    assert_node(
        doctree[3][1][0][0][1][0][2][0],
        (
            [addnodes.literal_strong, 'items'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'Tuple'],
            [addnodes.literal_emphasis, '['],
            [pending_xref, addnodes.literal_emphasis, 'str'],
            [addnodes.literal_emphasis, ', '],
            [addnodes.literal_emphasis, '...'],
            [addnodes.literal_emphasis, ']'],
            ')',
            ' -- ',
            'blah blah',
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][2][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='Tuple',
        **{'py:module': 'example', 'py:class': 'Class'},
    )
    assert_node(
        doctree[3][1][0][0][1][0][2][0][4],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='str',
        **{'py:module': 'example', 'py:class': 'Class'},
    )

    # :param Dict[str, str] params:
    assert_node(
        doctree[3][1][0][0][1][0][3][0],
        (
            [addnodes.literal_strong, 'params'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'Dict'],
            [addnodes.literal_emphasis, '['],
            [pending_xref, addnodes.literal_emphasis, 'str'],
            [addnodes.literal_emphasis, ', '],
            [pending_xref, addnodes.literal_emphasis, 'str'],
            [addnodes.literal_emphasis, ']'],
            ')',
            ' -- ',
            'blah blah',
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][3][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='Dict',
        **{'py:module': 'example', 'py:class': 'Class'},
    )
    assert_node(
        doctree[3][1][0][0][1][0][3][0][4],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='str',
        **{'py:module': 'example', 'py:class': 'Class'},
    )
    assert_node(
        doctree[3][1][0][0][1][0][3][0][6],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='str',
        **{'py:module': 'example', 'py:class': 'Class'},
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list_piped_type(app):
    text = (
        '.. py:module:: example\n'
        '.. py:class:: Class\n'
        '\n'
        '   :param age: blah blah\n'
        '   :type age: int | str\n'
    )
    doctree = restructuredtext.parse(app, text)

    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [
                        desc_content,
                        nodes.field_list,
                        nodes.field,
                        (nodes.field_name, nodes.field_body),
                    ],
                ),
            ],
        ),
    )
    assert_node(
        doctree[3][1][0][0][1],
        (
            [
                nodes.paragraph,
                (
                    [addnodes.literal_strong, 'age'],
                    ' (',
                    [pending_xref, addnodes.literal_emphasis, 'int'],
                    [addnodes.literal_emphasis, ' | '],
                    [pending_xref, addnodes.literal_emphasis, 'str'],
                    ')',
                    ' -- ',
                    'blah blah',
                ),
            ],
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='int',
        **{'py:module': 'example', 'py:class': 'Class'},
    )
    assert_node(
        doctree[3][1][0][0][1][0][4],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='str',
        **{'py:module': 'example', 'py:class': 'Class'},
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list_Literal(app):
    text = (
        '.. py:module:: example\n'
        '.. py:class:: Class\n'
        '\n'
        '   :param age: blah blah\n'
        "   :type age: Literal['foo', 'bar', 'baz']\n"
    )
    doctree = restructuredtext.parse(app, text)

    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [
                        desc_content,
                        nodes.field_list,
                        nodes.field,
                        (nodes.field_name, nodes.field_body),
                    ],
                ),
            ],
        ),
    )
    assert_node(
        doctree[3][1][0][0][1],
        (
            [
                nodes.paragraph,
                (
                    [addnodes.literal_strong, 'age'],
                    ' (',
                    [pending_xref, addnodes.literal_emphasis, 'Literal'],
                    [addnodes.literal_emphasis, '['],
                    [addnodes.literal_emphasis, "'foo'"],
                    [addnodes.literal_emphasis, ', '],
                    [addnodes.literal_emphasis, "'bar'"],
                    [addnodes.literal_emphasis, ', '],
                    [addnodes.literal_emphasis, "'baz'"],
                    [addnodes.literal_emphasis, ']'],
                    ')',
                    ' -- ',
                    'blah blah',
                ),
            ],
        ),
    )
    assert_node(
        doctree[3][1][0][0][1][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='Literal',
        **{'py:module': 'example', 'py:class': 'Class'},
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list_var(app):
    text = '.. py:class:: Class\n\n   :var int attr: blah blah\n'
    doctree = restructuredtext.parse(app, text)

    assert_node(
        doctree,
        (
            addnodes.index,
            [desc, (desc_signature, [desc_content, nodes.field_list, nodes.field])],
        ),
    )
    assert_node(
        doctree[1][1][0][0],
        ([nodes.field_name, 'Variables'], [nodes.field_body, nodes.paragraph]),
    )

    # :var int attr:
    assert_node(
        doctree[1][1][0][0][1][0],
        (
            [addnodes.literal_strong, 'attr'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'int'],
            ')',
            ' -- ',
            'blah blah',
        ),
    )
    assert_node(
        doctree[1][1][0][0][1][0][2],
        pending_xref,
        refdomain='py',
        reftype='class',
        reftarget='int',
        **{'py:class': 'Class'},
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list_napoleon_deliminator_of(app):
    text = (
        '.. py:module:: example\n'
        '.. py:class:: Class\n'
        '\n'
        '   :param list_str_var: example description.\n'
        '   :type list_str_var: list of str\n'
        '   :param tuple_int_var: example description.\n'
        '   :type tuple_int_var: tuple of tuple of int\n'
    )
    doctree = restructuredtext.parse(app, text)

    # :param list of str list_str_var:
    assert_node(
        doctree[3][1][0][0][1][0][0][0],
        (
            [addnodes.literal_strong, 'list_str_var'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'list'],
            [addnodes.literal_emphasis, ' of '],
            [pending_xref, addnodes.literal_emphasis, 'str'],
            ')',
            ' -- ',
            'example description.',
        ),
    )

    # :param tuple of tuple of int tuple_int_var:
    assert_node(
        doctree[3][1][0][0][1][0][1][0],
        (
            [addnodes.literal_strong, 'tuple_int_var'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'tuple'],
            [addnodes.literal_emphasis, ' of '],
            [pending_xref, addnodes.literal_emphasis, 'tuple'],
            [addnodes.literal_emphasis, ' of '],
            [pending_xref, addnodes.literal_emphasis, 'int'],
            ')',
            ' -- ',
            'example description.',
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_info_field_list_napoleon_deliminator_or(app):
    text = (
        '.. py:module:: example\n'
        '.. py:class:: Class\n'
        '\n'
        '   :param bool_str_var: example description.\n'
        '   :type bool_str_var: bool or str\n'
        '   :param str_float_int_var: example description.\n'
        '   :type str_float_int_var: str or float or int\n'
    )
    doctree = restructuredtext.parse(app, text)

    # :param bool or str bool_str_var:
    assert_node(
        doctree[3][1][0][0][1][0][0][0],
        (
            [addnodes.literal_strong, 'bool_str_var'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'bool'],
            [addnodes.literal_emphasis, ' or '],
            [pending_xref, addnodes.literal_emphasis, 'str'],
            ')',
            ' -- ',
            'example description.',
        ),
    )

    # :param str or float or int str_float_int_var:
    assert_node(
        doctree[3][1][0][0][1][0][1][0],
        (
            [addnodes.literal_strong, 'str_float_int_var'],
            ' (',
            [pending_xref, addnodes.literal_emphasis, 'str'],
            [addnodes.literal_emphasis, ' or '],
            [pending_xref, addnodes.literal_emphasis, 'float'],
            [addnodes.literal_emphasis, ' or '],
            [pending_xref, addnodes.literal_emphasis, 'int'],
            ')',
            ' -- ',
            'example description.',
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_type_field(app):
    text = (
        '.. py:data:: var1\n'
        '   :type: .int\n'
        '.. py:data:: var2\n'
        '   :type: ~builtins.int\n'
        '.. py:data:: var3\n'
        '   :type: typing.Optional[typing.Tuple[int, typing.Any]]\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'var1'],
                            [
                                desc_annotation,
                                (
                                    [desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, 'int'],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'var2'],
                            [
                                desc_annotation,
                                (
                                    [desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, 'int'],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'var3'],
                            [
                                desc_annotation,
                                (
                                    [desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, 'Optional'],
                                    [desc_sig_punctuation, '['],
                                    [pending_xref, 'Tuple'],
                                    [desc_sig_punctuation, '['],
                                    [pending_xref, 'int'],
                                    [desc_sig_punctuation, ','],
                                    desc_sig_space,
                                    [pending_xref, 'Any'],
                                    [desc_sig_punctuation, ']'],
                                    [desc_sig_punctuation, ']'],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(doctree[1][0][1][2], pending_xref, reftarget='int', refspecific=True)
    assert_node(
        doctree[3][0][1][2], pending_xref, reftarget='builtins.int', refspecific=False
    )
    assert_node(
        doctree[5][0][1][2],
        pending_xref,
        reftarget='typing.Optional',
        refspecific=False,
    )
    assert_node(
        doctree[5][0][1][4], pending_xref, reftarget='typing.Tuple', refspecific=False
    )
    assert_node(doctree[5][0][1][6], pending_xref, reftarget='int', refspecific=False)
    assert_node(
        doctree[5][0][1][9], pending_xref, reftarget='typing.Any', refspecific=False
    )



from autosummary_dummy_module import Foo


class InheritedAttrClass(Foo):
    def __init__(self):
        #: other docstring
        self.subclassattr = 'subclassattr'

        super().__init__()


__all__ = ['InheritedAttrClass']



"""Test sphinx.util.docstrings."""

from __future__ import annotations

from sphinx.util.docstrings import (
    prepare_commentdoc,
    prepare_docstring,
    separate_metadata,
)


def test_separate_metadata() -> None:
    # metadata only
    text = ':meta foo: bar\n:meta baz:\n'
    docstring, metadata = separate_metadata(text)
    assert docstring == ''
    assert metadata == {'foo': 'bar', 'baz': ''}

    # non metadata field list item
    text = ':meta foo: bar\n:param baz:\n'
    docstring, metadata = separate_metadata(text)
    assert docstring == ':param baz:\n'
    assert metadata == {'foo': 'bar'}

    # field_list like text following just after paragraph is not a field_list
    text = 'blah blah blah\n:meta foo: bar\n:meta baz:\n'
    docstring, metadata = separate_metadata(text)
    assert docstring == text
    assert metadata == {}

    # field_list like text following after blank line is a field_list
    text = 'blah blah blah\n\n:meta foo: bar\n:meta baz:\n'
    docstring, metadata = separate_metadata(text)
    assert docstring == 'blah blah blah\n\n'
    assert metadata == {'foo': 'bar', 'baz': ''}

    # non field_list item breaks field_list
    text = ':meta foo: bar\nblah blah blah\n:meta baz:\n'
    docstring, metadata = separate_metadata(text)
    assert docstring == 'blah blah blah\n:meta baz:\n'
    assert metadata == {'foo': 'bar'}


def test_prepare_docstring() -> None:
    docstring = """multiline docstring

                Lorem ipsum dolor sit amet, consectetur adipiscing elit,
                sed do eiusmod tempor incididunt ut labore et dolore magna
                aliqua::

                  Ut enim ad minim veniam, quis nostrud exercitation
                    ullamco laboris nisi ut aliquip ex ea commodo consequat.
                """

    assert prepare_docstring(docstring) == [
        'multiline docstring',
        '',
        'Lorem ipsum dolor sit amet, consectetur adipiscing elit,',
        'sed do eiusmod tempor incididunt ut labore et dolore magna',
        'aliqua::',
        '',
        '  Ut enim ad minim veniam, quis nostrud exercitation',
        '    ullamco laboris nisi ut aliquip ex ea commodo consequat.',
        '',
    ]

    docstring = """

                multiline docstring with leading empty lines
                """
    assert prepare_docstring(docstring) == [
        'multiline docstring with leading empty lines',
        '',
    ]

    docstring = 'single line docstring'
    assert prepare_docstring(docstring) == ['single line docstring', '']


def test_prepare_commentdoc() -> None:
    assert prepare_commentdoc('hello world') == []
    assert prepare_commentdoc('#: hello world') == ['hello world', '']
    assert prepare_commentdoc('#:  hello world') == [' hello world', '']
    assert prepare_commentdoc('#: hello\n#: world\n') == ['hello', 'world', '']



"""Test the build process with Texinfo builder with the test root."""

from __future__ import annotations

import re
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from sphinx.builders.texinfo import default_texinfo_documents
from sphinx.config import Config
from sphinx.util.docutils import new_document
from sphinx.writers.texinfo import TexinfoTranslator

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('texinfo', testroot='root')
def test_texinfo(app: SphinxTestApp) -> None:
    TexinfoTranslator.ignore_missing_images = True
    app.build(force_all=True)
    result = (app.outdir / 'sphinxtests.texi').read_text(encoding='utf8')
    assert (
        '@anchor{markup doc}@anchor{11}'
        '@anchor{markup id1}@anchor{12}'
        '@anchor{markup testing-various-markup}@anchor{13}'
    ) in result
    assert 'Footnotes' not in result
    # now, try to run makeinfo over it
    try:
        args = ['makeinfo', '--no-split', 'sphinxtests.texi']
        subprocess.run(args, capture_output=True, cwd=app.outdir, check=True)
    except OSError as exc:
        raise pytest.skip.Exception from exc  # most likely makeinfo was not found
    except CalledProcessError as exc:
        print(exc.stdout)
        print(exc.stderr)
        msg = f'makeinfo exited with return code {exc.returncode}'
        raise AssertionError(msg) from exc


@pytest.mark.sphinx('texinfo', testroot='markup-rubric')
def test_texinfo_rubric(app: SphinxTestApp) -> None:
    app.build()

    output = (app.outdir / 'projectnamenotset.texi').read_text(encoding='utf8')
    assert '@heading This is a rubric' in output
    assert '@heading This is a multiline rubric' in output
    assert '@heading A rubric with a heading level' in output


@pytest.mark.sphinx('texinfo', testroot='markup-citation')
def test_texinfo_citation(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    output = (app.outdir / 'projectnamenotset.texi').read_text(encoding='utf8')
    assert 'This is a citation ref; @ref{1,,[CITE1]} and @ref{2,,[CITE2]}.' in output
    assert (
        '@anchor{index cite1}@anchor{1}@w{(CITE1)} \nThis is a citation\n'
    ) in output
    assert (
        '@anchor{index cite2}@anchor{2}@w{(CITE2)} \nThis is a multiline citation\n'
    ) in output


def test_default_texinfo_documents() -> None:
    config = Config({
        'project': 'STASI™ Documentation',
        'author': "Wolfgang Schäuble & G'Beckstein",
    })
    expected = [
        (
            'index',
            'stasi',
            'STASI™ Documentation',
            "Wolfgang Schäuble & G'Beckstein",
            'stasi',
            'One line description of project',
            'Miscellaneous',
        )
    ]
    assert default_texinfo_documents(config) == expected


@pytest.mark.sphinx('texinfo', testroot='root')
def test_texinfo_escape_id(app: SphinxTestApp) -> None:
    settings = Mock(title='', texinfo_dir_entry='', texinfo_elements={})
    document = new_document('', settings)
    translator = app.builder.create_translator(document, app.builder)
    assert isinstance(translator, TexinfoTranslator)  # type-checking

    assert translator.escape_id('Hello world') == 'Hello world'
    assert translator.escape_id('Hello    world') == 'Hello world'
    assert translator.escape_id('Hello   Sphinx   world') == 'Hello Sphinx world'
    assert translator.escape_id('Hello:world') == 'Hello world'
    assert translator.escape_id('Hello(world)') == 'Hello world'
    assert translator.escape_id('Hello world.') == 'Hello world'
    assert translator.escape_id('.') == '.'


@pytest.mark.sphinx('texinfo', testroot='footnotes')
def test_texinfo_footnote(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    output = (app.outdir / 'projectnamenotset.texi').read_text(encoding='utf8')
    assert 'First footnote: @footnote{\nFirst\n}' in output


@pytest.mark.sphinx('texinfo', testroot='root')
def test_texinfo_xrefs(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    output = (app.outdir / 'sphinxtests.texi').read_text(encoding='utf8')
    assert re.search(r'@ref{\w+,,--plugin\.option}', output)

    # Now rebuild it without xrefs
    app.config.texinfo_cross_references = False
    app.build(force_all=True)
    output = (app.outdir / 'sphinxtests.texi').read_text(encoding='utf8')
    assert not re.search(r'@ref{\w+,,--plugin\.option}', output)
    assert (
        'Link to perl +p, --ObjC++, --plugin.option, create-auth-token, arg and -j'
    ) in output


@pytest.mark.sphinx('texinfo', testroot='root')
def test_texinfo_samp_with_variable(app: SphinxTestApp) -> None:
    app.build()

    output = (app.outdir / 'sphinxtests.texi').read_text(encoding='utf8')

    assert '@code{@var{variable_only}}' in output
    assert '@code{@var{variable} and text}' in output
    assert '@code{Show @var{variable} in the middle}' in output


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx('texinfo', testroot='images')
def test_copy_images(app: SphinxTestApp) -> None:
    app.build()

    images_dir = Path(app.outdir) / 'projectnamenotset-figures'
    images = {image.name for image in images_dir.rglob('*')}
    images.discard('python-logo.png')
    assert images == {
        'ba30773957c3fe046897111afd65a80b81cad089.png',  # texinfo: image from data:image/png URI in source
        'img.png',
        'rimg.png',
        'testimäge.png',
    }  # fmt: skip



"""Tests the Python Domain"""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_optional,
    desc_parameter,
    desc_parameterlist,
    desc_returns,
    desc_sig_keyword,
    desc_sig_name,
    desc_sig_operator,
    desc_sig_punctuation,
    desc_sig_space,
    desc_signature,
    pending_xref,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction(app):
    text = (
        '.. py:function:: func1\n'
        '.. py:module:: example\n'
        '.. py:function:: func2\n'
        '   :async:\n'
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_name, 'func1'], [desc_parameterlist, ()])],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'async'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'func2'],
                            [desc_parameterlist, ()],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'built-in function; func1()', 'func1', '', None)],
    )
    assert_node(
        doctree[2],
        addnodes.index,
        entries=[('pair', 'module; example', 'module-example', '', None)],
    )
    assert_node(
        doctree[3],
        addnodes.index,
        entries=[('single', 'func2() (in module example)', 'example.func2', '', None)],
    )

    assert 'func1' in domain.objects
    assert domain.objects['func1'] == ('index', 'func1', 'function', False)
    assert 'example.func2' in domain.objects
    assert domain.objects['example.func2'] == (
        'index',
        'example.func2',
        'function',
        False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_signature(app):
    text = '.. py:function:: hello(name: str) -> str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'name'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_signature_full(app):
    text = (
        '.. py:function:: hello(a: str, b = 1, *args: str, '
        'c: bool = True, d: tuple = (1, 2), **kwargs: str) -> str'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'a'],
                        [desc_sig_punctuation, ':'],
                        desc_sig_space,
                        [desc_sig_name, pending_xref, 'str'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'b'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '1'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_operator, '*'],
                        [desc_sig_name, 'args'],
                        [desc_sig_punctuation, ':'],
                        desc_sig_space,
                        [desc_sig_name, pending_xref, 'str'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'c'],
                        [desc_sig_punctuation, ':'],
                        desc_sig_space,
                        [desc_sig_name, pending_xref, 'bool'],
                        desc_sig_space,
                        [desc_sig_operator, '='],
                        desc_sig_space,
                        [nodes.inline, 'True'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'd'],
                        [desc_sig_punctuation, ':'],
                        desc_sig_space,
                        [desc_sig_name, pending_xref, 'tuple'],
                        desc_sig_space,
                        [desc_sig_operator, '='],
                        desc_sig_space,
                        [nodes.inline, '(1, 2)'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_operator, '**'],
                        [desc_sig_name, 'kwargs'],
                        [desc_sig_punctuation, ':'],
                        desc_sig_space,
                        [desc_sig_name, pending_xref, 'str'],
                    ),
                ],
            ),
        ],
    )
    # case: separator at head
    text = '.. py:function:: hello(*, a)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '*'],
                [desc_parameter, desc_sig_name, 'a'],
            ),
        ],
    )

    # case: separator in the middle
    text = '.. py:function:: hello(a, /, b, *, c)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [desc_parameter, desc_sig_name, 'a'],
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '/'],
                [desc_parameter, desc_sig_name, 'b'],
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '*'],
                [desc_parameter, desc_sig_name, 'c'],
            ),
        ],
    )

    # case: separator in the middle (2)
    text = '.. py:function:: hello(a, /, *, b)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [desc_parameter, desc_sig_name, 'a'],
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '/'],
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '*'],
                [desc_parameter, desc_sig_name, 'b'],
            ),
        ],
    )

    # case: separator at tail
    text = '.. py:function:: hello(a, /)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [desc_parameter, desc_sig_name, 'a'],
                [desc_parameter, desc_sig_operator, nodes.abbreviation, '/'],
            ),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_with_unary_operators(app):
    text = '.. py:function:: menu(egg=+1, bacon=-1, sausage=~1, spam=not spam)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'egg'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '+1'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'bacon'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '-1'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'sausage'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '~1'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'spam'],
                        [desc_sig_operator, '='],
                        [nodes.inline, 'not spam'],
                    ),
                ],
            ),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_with_binary_operators(app):
    text = '.. py:function:: menu(spam=2**64)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            ([
                desc_parameter,
                (
                    [desc_sig_name, 'spam'],
                    [desc_sig_operator, '='],
                    [nodes.inline, '2**64'],
                ),
            ]),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_with_number_literals(app):
    text = '.. py:function:: hello(age=0x10, height=1_6_0)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'age'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '0x10'],
                    ),
                ],
                [
                    desc_parameter,
                    (
                        [desc_sig_name, 'height'],
                        [desc_sig_operator, '='],
                        [nodes.inline, '1_6_0'],
                    ),
                ],
            ),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyfunction_with_union_type_operator(app):
    text = '.. py:function:: hello(age: int | None)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            ([
                desc_parameter,
                (
                    [desc_sig_name, 'age'],
                    [desc_sig_punctuation, ':'],
                    desc_sig_space,
                    [
                        desc_sig_name,
                        (
                            [pending_xref, 'int'],
                            desc_sig_space,
                            [desc_sig_punctuation, '|'],
                            desc_sig_space,
                            [pending_xref, 'None'],
                        ),
                    ],
                ),
            ]),
        ],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_optional_pyfunction_signature(app):
    text = '.. py:function:: compile(source [, filename [, symbol]]) -> ast object'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'compile'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'ast object'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        (
            [desc_parameter, ([desc_sig_name, 'source'])],
            [
                desc_optional,
                (
                    [desc_parameter, ([desc_sig_name, 'filename'])],
                    [desc_optional, desc_parameter, ([desc_sig_name, 'symbol'])],
                ),
            ],
        ),
    )


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'python_maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_python_maximum_signature_line_length_equal(app):
    text = '.. py:function:: hello(name: str) -> str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'name'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'python_maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_python_maximum_signature_line_length_force_single(
    app,
):
    text = '.. py:function:: hello(names: str) -> str\n   :single-line-parameter-list:'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'names'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'python_maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_python_maximum_signature_line_length_break(app):
    text = '.. py:function:: hello(names: str) -> str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'names'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=True)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_maximum_signature_line_length_equal(app):
    text = '.. py:function:: hello(name: str) -> str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'name'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_maximum_signature_line_length_force_single(app):
    text = '.. py:function:: hello(names: str) -> str\n   :single-line-parameter-list:'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'names'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name: str) -> str'),
    },
)
def test_pyfunction_signature_with_maximum_signature_line_length_break(app):
    text = '.. py:function:: hello(names: str) -> str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'hello'],
                            desc_parameterlist,
                            [desc_returns, pending_xref, 'str'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            desc_parameter,
            (
                [desc_sig_name, 'names'],
                [desc_sig_punctuation, ':'],
                desc_sig_space,
                [nodes.inline, pending_xref, 'str'],
            ),
        ],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=True)



import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().resolve() / 'src'))

extensions = ['sphinx.ext.apidoc']

apidoc_include_private = True
apidoc_follow_links = False
apidoc_separate_modules = True
apidoc_no_headings = False
apidoc_module_first = True
apidoc_modules = [
    {
        'path': 'src',
        'destination': 'generated',
        'exclude_patterns': ['src/exclude_package.py'],
        'automodule_options': ['members', 'undoc-members'],
        'max_depth': 3,
        'implicit_namespaces': False,
    }
]



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest

from tests.test_builders.xpath_data import FIGURE_CAPTION
from tests.test_builders.xpath_util import check_xpath

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='numfig')
@pytest.mark.test_params(shared_result='test_build_html_numfig')
def test_numfig_disabled_warn(app: SphinxTestApp) -> None:
    app.build()
    warnings = app.warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' not in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' not in warnings


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
        ('index.html', ".//table/caption/span[@class='caption-number']", None, True),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            None,
            True,
        ),
        ('index.html', './/li/p/code/span', '^fig1$', True),
        ('index.html', './/li/p/code/span', '^Figure%s$', True),
        ('index.html', './/li/p/code/span', '^table-1$', True),
        ('index.html', './/li/p/code/span', '^Table:%s$', True),
        ('index.html', './/li/p/code/span', '^CODE_1$', True),
        ('index.html', './/li/p/code/span', '^Code-%s$', True),
        ('index.html', './/li/p/a/span', '^Section 1$', True),
        ('index.html', './/li/p/a/span', '^Section 2.1$', True),
        ('index.html', './/li/p/code/span', '^Fig.{number}$', True),
        ('index.html', './/li/p/a/span', '^Sect.1 Foo$', True),
        ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
        ('foo.html', ".//table/caption/span[@class='caption-number']", None, True),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            None,
            True,
        ),
        ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
        ('bar.html', ".//table/caption/span[@class='caption-number']", None, True),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            None,
            True,
        ),
        ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
        ('baz.html', ".//table/caption/span[@class='caption-number']", None, True),
        (
            'baz.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            None,
            True,
        ),
    ],
)
@pytest.mark.sphinx('html', testroot='numfig')
@pytest.mark.test_params(shared_result='test_build_html_numfig')
def test_numfig_disabled(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    srcdir='test_numfig_without_numbered_toctree_warn',
    confoverrides={'numfig': True},
)
def test_numfig_without_numbered_toctree_warn(app: SphinxTestApp) -> None:
    app.build()
    # remove :numbered: option
    index = (app.srcdir / 'index.rst').read_text(encoding='utf8')
    index = re.sub(':numbered:.*', '', index)
    (app.srcdir / 'index.rst').write_text(index, encoding='utf8')
    app.build()

    warnings = app.warning.getvalue()
    assert (
        'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.'
    ) not in warnings
    assert (
        'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index'
    ) in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 9 $',
            True,
        ),
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 10 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 9 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 10 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 9 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 10 $',
            True,
        ),
        ('index.html', './/li/p/a/span', '^Fig. 9$', True),
        ('index.html', './/li/p/a/span', '^Figure6$', True),
        ('index.html', './/li/p/a/span', '^Table 9$', True),
        ('index.html', './/li/p/a/span', '^Table:6$', True),
        ('index.html', './/li/p/a/span', '^Listing 9$', True),
        ('index.html', './/li/p/a/span', '^Code-6$', True),
        ('index.html', './/li/p/code/span', '^foo$', True),
        ('index.html', './/li/p/code/span', '^bar_a$', True),
        ('index.html', './/li/p/a/span', '^Fig.9 should be Fig.1$', True),
        ('index.html', './/li/p/code/span', '^Sect.{number}$', True),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 3 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 4 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 3 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 4 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 3 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 4 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 5 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 7 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 8 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 5 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 7 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 8 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 5 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 7 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 8 $',
            True,
        ),
        (
            'baz.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 6 $',
            True,
        ),
        (
            'baz.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 6 $',
            True,
        ),
        (
            'baz.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 6 $',
            True,
        ),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    srcdir='test_numfig_without_numbered_toctree',
    confoverrides={'numfig': True},
)
def test_numfig_without_numbered_toctree(
    app, cached_etree_parse, fname, path, check, be_found
):
    # remove :numbered: option
    index = (app.srcdir / 'index.rst').read_text(encoding='utf8')
    index = re.sub(':numbered:.*', '', index)
    (app.srcdir / 'index.rst').write_text(index, encoding='utf8')

    if not list(app.outdir.iterdir()):
        app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={'numfig': True},
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_numbered_toctree_warn(app: SphinxTestApp) -> None:
    app.build()
    warnings = app.warning.getvalue()
    assert (
        'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.'
    ) not in warnings
    assert (
        'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index'
    ) in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1 $',
            True,
        ),
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2 $',
            True,
        ),
        ('index.html', './/li/p/a/span', '^Fig. 1$', True),
        ('index.html', './/li/p/a/span', '^Figure2.2$', True),
        ('index.html', './/li/p/a/span', '^Table 1$', True),
        ('index.html', './/li/p/a/span', '^Table:2.2$', True),
        ('index.html', './/li/p/a/span', '^Listing 1$', True),
        ('index.html', './/li/p/a/span', '^Code-2.2$', True),
        ('index.html', './/li/p/a/span', '^Section.1$', True),
        ('index.html', './/li/p/a/span', '^Section.2.1$', True),
        ('index.html', './/li/p/a/span', '^Fig.1 should be Fig.1$', True),
        ('index.html', './/li/p/a/span', '^Sect.1 Foo$', True),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.1 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.2 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.3 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.4 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.3 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.4 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.3 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.4 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.1 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.3 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.4 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.3 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.4 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.3 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.4 $',
            True,
        ),
        (
            'baz.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.2 $',
            True,
        ),
        (
            'baz.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.2 $',
            True,
        ),
        (
            'baz.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.2 $',
            True,
        ),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={'numfig': True},
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_numbered_toctree(
    app, cached_etree_parse, fname, path, check, be_found
):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={
        'numfig': True,
        'numfig_format': {
            'figure': 'Figure:%s',
            'table': 'Tab_%s',
            'code-block': 'Code-%s',
            'section': 'SECTION-%s',
        },
    },
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_format_warn')
def test_numfig_with_prefix_warn(app: SphinxTestApp) -> None:
    app.build()
    warnings = app.warning.getvalue()
    assert (
        'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.'
    ) not in warnings
    assert (
        'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index'
    ) in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:1 $',
            True,
        ),
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:2 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_1 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_2 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-1 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-2 $',
            True,
        ),
        ('index.html', './/li/p/a/span', '^Figure:1$', True),
        ('index.html', './/li/p/a/span', '^Figure2.2$', True),
        ('index.html', './/li/p/a/span', '^Tab_1$', True),
        ('index.html', './/li/p/a/span', '^Table:2.2$', True),
        ('index.html', './/li/p/a/span', '^Code-1$', True),
        ('index.html', './/li/p/a/span', '^Code-2.2$', True),
        ('index.html', './/li/p/a/span', '^SECTION-1$', True),
        ('index.html', './/li/p/a/span', '^SECTION-2.1$', True),
        ('index.html', './/li/p/a/span', '^Fig.1 should be Fig.1$', True),
        ('index.html', './/li/p/a/span', '^Sect.1 Foo$', True),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:1.1 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:1.2 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:1.3 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:1.4 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_1.3 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_1.4 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-1.3 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-1.4 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:2.1 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:2.3 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:2.4 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_2.3 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_2.4 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-2.3 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-2.4 $',
            True,
        ),
        (
            'baz.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Figure:2.2 $',
            True,
        ),
        (
            'baz.html',
            ".//table/caption/span[@class='caption-number']",
            '^Tab_2.2 $',
            True,
        ),
        (
            'baz.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Code-2.2 $',
            True,
        ),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={
        'numfig': True,
        'numfig_format': {
            'figure': 'Figure:%s',
            'table': 'Tab_%s',
            'code-block': 'Code-%s',
            'section': 'SECTION-%s',
        },
    },
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_format_warn')
def test_numfig_with_prefix(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={'numfig': True, 'numfig_secnum_depth': 2},
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_depth_2')
def test_numfig_with_secnum_depth_warn(app: SphinxTestApp) -> None:
    app.build()
    warnings = app.warning.getvalue()
    assert (
        'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.'
    ) not in warnings
    assert (
        'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index'
    ) in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1 $',
            True,
        ),
        (
            'index.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1 $',
            True,
        ),
        (
            'index.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1 $',
            True,
        ),
        (
            'index.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2 $',
            True,
        ),
        ('index.html', './/li/p/a/span', '^Fig. 1$', True),
        ('index.html', './/li/p/a/span', '^Figure2.1.2$', True),
        ('index.html', './/li/p/a/span', '^Table 1$', True),
        ('index.html', './/li/p/a/span', '^Table:2.1.2$', True),
        ('index.html', './/li/p/a/span', '^Listing 1$', True),
        ('index.html', './/li/p/a/span', '^Code-2.1.2$', True),
        ('index.html', './/li/p/a/span', '^Section.1$', True),
        ('index.html', './/li/p/a/span', '^Section.2.1$', True),
        ('index.html', './/li/p/a/span', '^Fig.1 should be Fig.1$', True),
        ('index.html', './/li/p/a/span', '^Sect.1 Foo$', True),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.1 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.1.1 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.1.2 $',
            True,
        ),
        (
            'foo.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 1.2.1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 1.2.1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.1.1 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.1.2 $',
            True,
        ),
        (
            'foo.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.2.1 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.1.1 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.1.3 $',
            True,
        ),
        (
            'bar.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.1.1 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.1.3 $',
            True,
        ),
        (
            'bar.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.2.1 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.1.1 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.1.3 $',
            True,
        ),
        (
            'bar.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.2.1 $',
            True,
        ),
        (
            'baz.html',
            FIGURE_CAPTION + "/span[@class='caption-number']",
            '^Fig. 2.1.2 $',
            True,
        ),
        (
            'baz.html',
            ".//table/caption/span[@class='caption-number']",
            '^Table 2.1.2 $',
            True,
        ),
        (
            'baz.html',
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.1.2 $',
            True,
        ),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='numfig',
    confoverrides={'numfig': True, 'numfig_secnum_depth': 2},
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_depth_2')
def test_numfig_with_secnum_depth(
    app, cached_etree_parse, fname, path, check, be_found
):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.parametrize(
    'expect',
    [
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 1 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 2 $', True),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2 $',
            True,
        ),
        ('.//li/p/a/span', '^Fig. 1$', True),
        ('.//li/p/a/span', '^Figure2.2$', True),
        ('.//li/p/a/span', '^Table 1$', True),
        ('.//li/p/a/span', '^Table:2.2$', True),
        ('.//li/p/a/span', '^Listing 1$', True),
        ('.//li/p/a/span', '^Code-2.2$', True),
        ('.//li/p/a/span', '^Section.1$', True),
        ('.//li/p/a/span', '^Section.2.1$', True),
        ('.//li/p/a/span', '^Fig.1 should be Fig.1$', True),
        ('.//li/p/a/span', '^Sect.1 Foo$', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.2 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.3 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.4 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 1.1 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 1.2 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 1.3 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 1.4 $', True),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.1 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.2 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.3 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 1.4 $',
            True,
        ),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.3 $', True),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.4 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 2.1 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 2.3 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 2.4 $', True),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.1 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.3 $',
            True,
        ),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.4 $',
            True,
        ),
        (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.2 $', True),
        (".//table/caption/span[@class='caption-number']", '^Table 2.2 $', True),
        (
            ".//div[@class='code-block-caption']/span[@class='caption-number']",
            '^Listing 2.2 $',
            True,
        ),
    ],
)
@pytest.mark.sphinx(
    'singlehtml',
    testroot='numfig',
    confoverrides={'numfig': True},
)
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_singlehtml(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)
